# zipup.sh:
# Script to generate the distributions.
#
# By Christer Sandberg (ported from Adime by Sven Sandberg)
#
# Requires zip, tar, gzip, rm, cp, and sed.
#
# whisped (my own tool for converting tabs, newline characters and the like)
# will be used if it is available, but it's no big deal if this is not the
# case.
##

# note: these are patched by fixver.sh script.
zipversion=095
tgzversion=0.9.5

if [ "$1" != "-q" ]; then

echo clean up...
# Make unix line endings
./fix.sh unix --dtou

# Dont generate dependencies if a parameter was given.
# This is because I often realize that I missed something essential _after_
# calling zipup.sh, so it saves lots of time if the dependencies don't need
# to be rebuilt
if [ "$1" = "--no-depend" ]; then
   echo skip generating dependencies
else
   echo generate dll stuff...
   ./misc/fixdll.sh

   echo generate dependencies...
   #But don't alter the line endings (--quick ensures that)
   ./fix.sh mingw32 --quick
   make depend

   ./fix.sh msvc --quick
   make depend

   ./fix.sh unix --quick
   make depend

   ./fix.sh djgpp --quick
   make depend
fi

echo generate other files for the distribution...
make docs "LFN=y"
make dat

fi

echo copy default makefile...
cp misc/emptymak makefile

files="lib/*/tmpfile.txt \
       makefile misc/makefile* misc/emptymak misc/scanexp.c \
       fix* misc/*.sh \
       obj/*/*/*.dep lib/*/*.def \
       readme.txt readme.html docs/cpri.html docs/cpri.txt \
       resource/cprilabl resource/icons.dat \
       docs/*._tx \
       src/*.c src/*.h src/*.inc src/*.t src/*.ht \
       examples/*.c examples/*.cfg \
       resource/* \
       include/*.h include/cpri/*.h"

afiles=`echo $files       | sed -e"s/\([^ ][^ ]*\)/ cpri\/\1/g"`


#############################################################################

echo remove old archives...
rm -f cpri$zipversion.zip
rm -f cpri-$tgzversion.tar.gz

# Line endings are still unix in all files, including bat-files. To make the tar archive useful also for
# Windows users, we change the line endings to DOS/Windows already before packing the tar. This should be
# no problem for unix users since the bat file has no meaning in unix.
cp "fix.bat" _tmpfile
perl -p -i -e "s/([^\r]|^)\n/\1\r\n/" _tmpfile
touch -r "fix.bat" _tmpfile
mv _tmpfile "fix.bat"

echo create tar.gz archives...
cd ..
tar -cv $afiles | gzip -9  > cpri/cpri-$tgzversion.tar.gz
cd cpri

echo create zip archives...
# Change line endings from unix to DOS/Windows before packing the zip archives
./fix.sh djgpp --utod
cd ..
zip -9 cpri/cpri$zipversion.zip $afiles
cd cpri

echo "Archives finished!"

# Restore correct unix line endings
./fix.sh unix --dtou

# fix.sh omits to change line endings in the bat-file, so we need to change it separately because we want the local
# copy to have the unix style to have the right endings when committing to cvs next time.
mv "fix.bat" _tmpfile
tr -d '\015' < _tmpfile > "fix.bat"
touch -r _tmpfile "fix.bat"
rm _tmpfile

echo all done!
