#if DEGUBMODE==1 || defined _DEBUG
#define USE_LINES 0
#define USE_STD_PRINTER 0
#define DEBUG_PRINTS 1
#else 
#define USE_LINES 0
#define USE_STD_PRINTER 1
#define DEBUG_PRINTS 0
#endif
#if USE_LINES == 1
#define WINVER 0x0500
#define _WIN32_WINNT 0x0500
#endif

/* Module DRV_STDWIN.C
Contains the print-driver for windows-printers */
#ifdef _WIN32

#define DRV_OWN_GETSUBSEQUENTTEXT
#define DRV_OWN_GETPARAM
#define DRV_OWN_GETFLOATPARAM

#define VERT_TEXT_ALIGN TA_BASELINE
#include <windows.h>
#include <cgui/mem.h>

#include <string.h>

#include "cpri.h"
#include "spooler.h"
#include "pridrive.h"

#include "drv_common.h"
#include "drv_stdwin.h"
#include "width.h"
#include "dialpri.h"
#include "cpri.ht"

#if DEBUG_PRINTS==0
#ifdef _MSC_VER
#define dprintf
#else
#define dprintf(...)
#endif
#else
#define dprintf printf
#endif //DEBUG_PRINTS

#define FIX_SPACING 1
#define PROP_SPACING 0

#define SUP_NO_TEXT			0
#define SUP_TEXT_TO_LEFT	1
#define SUP_TEXT_TO_RIGHT	2

#define MAXROWLENGTH 128
#define POSSTACKSIZE 10

#define NBARWIDTH (5*0.084)
#define WBARWIDTH(nbarwidth) (int)(2.6*(nbarwidth)+0.5)
#define BARCODEHEIGHT 7
#define IndexSize(size) ( (size)*2/3 )
#define ROW_HEIGHT(dd) ((int)((dd->size/2.834645)*dd->scaling+0.5))

#define SCALING         300

typedef struct t_macro
{
	int id;                 //Macro number
	int start;
	int length;
	struct t_macro *next;
}t_macro;

typedef struct t_font
{
	int bottom2baseline;
	int baseline2top;
	int size;                    /* current font size */
	unsigned char style;         /* current font style */
	unsigned char stroke;        /* current stroke weight of font */
	int face;                    /* current font typeface */
	int spacing;                 /* current spacing: 0=PROP, 1=FIX */
	int strike;
	int ul;                      /* underline indicator */

}t_font;

typedef struct t_format
{
	t_tab_frame tf;              /* A list of the current tabs (in three different scalings) */
	t_tab margin;                /* Specifies the current margins. */
	int gridon;                  /* It is necessary to set gridon after	defining the set of tabs to be used */
	t_font ActiveFont,			 /* The font being printed */
		RowFont;				 /* The font specifying rowheiht */
}t_format;

typedef struct t_driver {
	int topmargin;
	int bottom;                  /* "soft" bottom margin */
	int minbottom;
	float xscaling,yscaling;               /* Scaling factor (currently values will be scaled from mm to suitabble values for the printer) */
	int rowheight;               /* The heigth of each line used for demand	line feed */
	int heightpercent;            /* The row-heigth expressed in % of standard height 4.2mm/12dpi font */
	int ypos;                    /* current printing y-position */
	int xpos;                    /* current printing x-position */
	int tablemode;				 /* Indicates if table mode is used */
	int linebreaking;			 /* Indicates that a rowbreaking is in progress */

	t_macro head;				// The head macro...
	int nocopies;                /* number of copies of each printout */
	int copies;                  /* flag indicationg that printing of copies is in progress */
	int pagelen;                 /* current page length (depending on orientation) */
	int pagewidth;
	char *nspace;                /* different barcodes for different printers */
	char *wspace;
	char *wbar;

	t_format ActiveFormat,
		HeadFormat;

	int IndexWaiting;
	char *IndexedText;

	char nextRow[MAXROWLENGTH+1];       /* Row to be printed */
	char *NextToken;              /* Next token to insert into nextRow*/

	//int printing_embedded;

	DOCINFO di;
	HDC ghdc;

	t_macro *macros;

	char *prnbuf;
	int pos;
	int bufsize;

	int CurrentMacro;

	char *outText;

	DEVMODE *dvmd;
	HANDLE hdvmd;
	int posStack[POSSTACKSIZE*2];
	int *posStackPos;

	struct t_printerq *q;

	int xoffset,yoffset;

	int barw,barh;				// Barcode with,height

	HFONT OrgFnt;

} t_driver;

static void PlayMacro(t_printerq *q,int id,t_macro *m);
static void driver(t_printerq *q);
static char *GetSubsequentText(t_driver * dd,  char *s);
static void FinnishLineBreak(t_driver *dd);
static void DrawRect(HDC ghdc,int,int,int,int);
static void IndexUpFromCurrentPos(t_driver *dd,char *s);

static void PlayHeadMacro(t_printerq *q)
{
	t_format NormalFormat;
	t_driver *dd=q->drv;

#if USE_LINES==1
	{
		HGDIOBJ pen;
		pen=SelectObject(dd->ghdc,GetStockObject(DC_PEN));
		SetDCPenColor(dd->ghdc,0x00FF0000);
		DrawRect(dd->ghdc,dd->ActiveFormat.margin.scaled-dd->xoffset,dd->topmargin,dd->pagewidth,dd->pagelen-dd->bottom);
		SelectObject(dd->ghdc,pen);
	}
#endif

	NormalFormat=dd->ActiveFormat;
	dd->ActiveFormat=dd->HeadFormat;
	dd->ActiveFormat.tf.tabs=NULL;
	_CpriCopyTabFrame(&dd->HeadFormat.tf,&dd->ActiveFormat.tf);

	PlayMacro(q,0,&dd->head);
	
	Release(dd->ActiveFormat.tf.tabs);
	dd->ActiveFormat=NormalFormat;

	//NewLine(dd);

}

static void ClearTabs(t_tab_frame *tf)
{
	tf->n=0;
}

static void printStr(t_driver *dd)
{
	POINT point;
	int x,y;
#if USE_LINES == 1
	POINT pts[2];
	HGDIOBJ pen;
#endif
	if(!GetTextAlign(dd->ghdc))
		SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
	*dd->NextToken='\0';
	if (*dd->outText)
	{
		if(dd->linebreaking)
			FinnishLineBreak(dd);

		x=dd->xpos;
		y=dd->ypos;
#if USE_LINES == 1
		pts[0].x=x;
		pts[0].y=y;
#endif
		if(dd->IndexWaiting!=SUP_NO_TEXT)
		{
			SIZE strSize;
			int CAlign;
			int xposLeft=0,xposRight=0;

			GetTextExtentPoint32(dd->ghdc,dd->outText,strlen(dd->outText),&strSize);
			CAlign=GetTextAlign(dd->ghdc);

			switch(CAlign&(TA_RIGHT|TA_LEFT|TA_CENTER))
			{
			case TA_LEFT:
				xposLeft=dd->xpos;
				xposRight=dd->xpos+strSize.cx;
				break;
			case TA_RIGHT:
				xposLeft=dd->xpos-strSize.cx;
				xposRight=dd->xpos;
				break;
			case TA_CENTER:
				xposLeft=dd->xpos-strSize.cx/2;
				xposRight=dd->xpos+strSize.cx/2;
				break;
			}
			if(dd->IndexWaiting==SUP_TEXT_TO_LEFT)
			{
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_RIGHT|TA_UPDATECP);
				dd->xpos=xposLeft;
			}
			else
			{
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
				dd->xpos=xposRight;
			}

			dd->IndexWaiting=SUP_NO_TEXT;
			IndexUpFromCurrentPos(dd,dd->IndexedText);

			SetTextAlign(dd->ghdc,CAlign);
			Release(dd->IndexedText);
		}
		MoveToEx(dd->ghdc,x,y,NULL);
		TextOut(dd->ghdc,x,y,dd->outText,strlen(dd->outText));
		
		dprintf("Printing %d chars @ (%d,%d): \"%s\"\n",strlen(dd->outText),x,y,dd->outText);
		dprintf("R_print '%s'\n",dd->outText);
		
		GetCurrentPositionEx(dd->ghdc,&point);

		dd->xpos=point.x;
		dd->ypos=point.y;
#if USE_LINES == 1
		pts[1]=point;
		pen=SelectObject(dd->ghdc,GetStockObject(DC_PEN));
		SetDCPenColor(dd->ghdc,0x00FF0000);
		Polyline(dd->ghdc,pts,2);
		SelectObject(dd->ghdc,pen);
#endif

		*dd->outText='\0';
		dd->NextToken=dd->outText;


	}
}

static void DrawRect(HDC ghdc,int x1,int y1,int x2,int y2)
{
	POINT pts[5];
	
	dprintf("Drawing rect from (%d,%d) to (%d,%d)\n",x1,y1,x2,y2);
	pts[0].x=x1;
	pts[0].y=y1;
	pts[4]=pts[3]=pts[2]=pts[1]=pts[0];
	pts[2].x=pts[1].x=x2;
	pts[3].y=pts[2].y=y2;
	
	Polyline(ghdc,pts,5);
}
static void MakeFrame(t_driver *dd)
{
	int i;
	int x1,x2,y1,y2;
#if USE_LINES == 1
	HGDIOBJ pen;
#endif
	
	x1=dd->ActiveFormat.tf.tabs[0].scaled-dd->xoffset;
	if(x1<0)
		x1=0;
	y1=dd->ypos-dd->ActiveFormat.RowFont.baseline2top;
	y2=dd->ypos+dd->ActiveFormat.RowFont.bottom2baseline;
#if USE_LINES ==1
	pen=SelectObject(dd->ghdc,GetStockObject(DC_PEN));
	SetDCPenColor(dd->ghdc,0x00FF0000);
	DrawRect(dd->ghdc,dd->ActiveFormat.tf.tabs[0].scaled-dd->xoffset,y1+20,dd->ActiveFormat.tf.tabs[0].scaled-dd->xoffset,y2-20);
	SelectObject(dd->ghdc,pen);
#endif
	for(i=1;i<dd->ActiveFormat.tf.n;i++)
	{
		if(dd->ActiveFormat.tf.tabs[i].inactive==0)
		{
			x2=dd->ActiveFormat.tf.tabs[i].scaled-dd->xoffset;
			DrawRect(dd->ghdc,x1,y1,x2,y2);
#if USE_LINES ==1
			pen=SelectObject(dd->ghdc,GetStockObject(DC_PEN));
			SetDCPenColor(dd->ghdc,0x00FF0000);
			DrawRect(dd->ghdc,dd->ActiveFormat.tf.tabs[i].scaled-dd->xoffset,y1+20,dd->ActiveFormat.tf.tabs[i].scaled-dd->xoffset,y2-20);
			SelectObject(dd->ghdc,pen);
#endif
			x1=x2;
		}
	}
	
}

static void NewPage(t_driver *dd)
{
	EndPage(dd->ghdc);
	StartPage(dd->ghdc);
	dprintf("<<NEW PAGE >>\n");
	if(dd->tablemode)
	{
		dd->ypos=dd->topmargin;
		if(dd->ypos<0)
			dd->ypos=0;
		PlayHeadMacro(dd->q);
	}
	else
	{
		dd->ypos=0;
	}
}

static void NewLine(t_driver *dd)
{
	printStr(dd);
	if(dd->ActiveFormat.gridon)
		MakeFrame(dd);
	dd->ActiveFormat.tf.curtab = 0;
	if(dd->tablemode)
		dd->xpos=dd->ActiveFormat.margin.scaled - dd->xoffset;
	else
		dd->xpos=0;
	if(dd->xpos<0)
		dd->xpos=0;
	dd->ypos+=dd->ActiveFormat.RowFont.bottom2baseline;
	dd->linebreaking=1;
	dprintf("Printing <<NEW LINE>>\n");
}

static void FinnishLineBreak(t_driver *dd)
{
	dd->ActiveFormat.RowFont=dd->ActiveFormat.ActiveFont;
	dd->rowheight=dd->ActiveFormat.RowFont.baseline2top+dd->ActiveFormat.RowFont.bottom2baseline;
	if(dd->ypos+dd->rowheight>dd->pagelen-dd->bottom)
		NewPage(dd);
	dd->ypos+=dd->ActiveFormat.RowFont.baseline2top;
	dd->linebreaking=0;

}

static void GetRes(t_driver *dd)
{
	dd->xscaling=(float)GetDeviceCaps(dd->ghdc,HORZRES)/GetDeviceCaps(dd->ghdc,HORZSIZE);
	dd->yscaling=(float)GetDeviceCaps(dd->ghdc,VERTRES)/GetDeviceCaps(dd->ghdc,VERTSIZE);
	dd->pagelen=GetDeviceCaps(dd->ghdc,PHYSICALHEIGHT);
	dd->xoffset=GetDeviceCaps(dd->ghdc,PHYSICALOFFSETX);
	dd->yoffset=GetDeviceCaps(dd->ghdc,PHYSICALOFFSETY);
	dd->pagewidth=GetDeviceCaps(dd->ghdc,HORZRES);
	dd->minbottom=dd->pagelen-(GetDeviceCaps(dd->ghdc,VERTRES)+dd->yoffset);
}

static int GetParam(t_printerq * q)
{
	int c, t, sgn = 1;
	
	t = 0;
	do {
		c = q->drv->prnbuf[q->drv->pos++];
		if (c == '-') {
			sgn = -1;
			c = q->drv->prnbuf[q->drv->pos++];
			break;
		}
	} while (c == ' ');
	while (q->drv->bufsize>q->drv->pos && c != 033)
	{
		t *= 10;
		t += c - 48;
		c = q->drv->prnbuf[q->drv->pos++];
	}
	return t * sgn;
}

static double GetFloatParam(t_printerq * q)
{
	char s[100], c;
	int i = 0;
	double f;
	
	
	do 
	{
		c = q->drv->prnbuf[q->drv->pos++];
		s[i++] = c;
	} while (q->drv->bufsize>q->drv->pos && c != 033);
	s[i - 1] = 0;
	sscanf(s, "%lf", &f);
	return f;
}

static char *GetSubsequentText(t_driver *dd, char *s)
{
	char c;
	
	do 
	{
		c = dd->prnbuf[dd->pos++];
		*s++ = c;
	} while (dd->bufsize>dd->pos && c != 033);
	s--;
	*s = 0;
	return s;
}

static void doRecordMacro(t_driver *dd,t_macro *m)
{
	m->start=dd->pos;
	dprintf("Start recording macro %d @ pos %d\n",m->id,dd->pos);
	for(;!(dd->prnbuf[dd->pos]==27 && (dd->prnbuf[dd->pos+1]==EPC_STOP_RECORD_MACRO ||dd->prnbuf[dd->pos+1]==EPC_END_TO_BE_REPEATED));dd->pos++)
		;
	dprintf("End macro recording @ %d\n",dd->pos);
	m->length=dd->pos;
	dd->pos+=2;
}

static void RecordMacro(t_driver *dd,int id,t_macro **m)
{
	if(*m)
		if((*m)->id==id)
			doRecordMacro(dd,*m);
		else
			RecordMacro(dd,id,&(*m)->next);
		else
		{
			*m=GetMem(t_macro,1);
			(*m)->next=NULL;
			(*m)->id=id;
			doRecordMacro(dd,*m);
		}
		
}

static void PlayMacro(t_printerq *q,int id,t_macro *m)
{
	int old_pos,old_bufsize;
	if(m) {
		if(m->id==id)
		{
			dprintf("Playing macro %d\n",id);
			old_pos=q->drv->pos;
			old_bufsize=q->drv->bufsize;
			q->drv->pos=m->start;
			q->drv->bufsize=m->length;
			driver(q);
			dprintf("End playing macro %d\n",id);
			q->drv->pos=old_pos;
			q->drv->bufsize=old_bufsize;
		}
		else
			PlayMacro(q,id,m->next);
	}
	
}

static void deleteMacros(t_macro **m)
{
	if(*m)
	{
		deleteMacros(&(*m)->next);
		Release(*m);
		*m=NULL;
	}
}


static void print_font(t_driver *dd)
{
	int facecode ;
	unsigned char spacing, family;
	LOGFONT lf;
	HFONT fnt;
	int face=dd->ActiveFormat.ActiveFont.face;
	unsigned char style=dd->ActiveFormat.ActiveFont.style;
	unsigned char stroke=dd->ActiveFormat.ActiveFont.stroke;
	TEXTMETRIC tm;
	
	family=FF_DONTCARE;
	spacing = PROP_SPACING;
	switch (face) 
	{
	case ALBERTUS:
		facecode = 0;
		style = 0;
		break;
	case ANTIQUE:
		facecode = 72;
		if (stroke && style)
			style = 0;
		break;
	case TIMES:
		facecode = 5;
		family=FF_ROMAN;
		break;
	case UNIVERS:
		facecode = 52;
		break;
	case CONDENSED:
		facecode = 52;
		break;
	case FONT5:
		facecode = 17;
		break;
	case COURIER:
		facecode = 3;
		spacing = FIX_SPACING;
		family=FF_MODERN;
		break;
	case GOTHIC:
		facecode = 6;
		spacing = FIX_SPACING;
		break;
	case SCRIPT:
		facecode = 20;
		style = 1;
		stroke = 0;
		family=FF_SCRIPT;
		break;
	case FONT10:
		facecode = 0;
		style = 1;
		stroke = 0;
		break;
	case FONT11:
		facecode = 0;
		style = 0;
		stroke = 1;
	default:
		facecode = 0;
		break;
	}
	lf.lfHeight=-MulDiv(dd->ActiveFormat.ActiveFont.size, GetDeviceCaps(dd->ghdc, LOGPIXELSY), 72);
	lf.lfWidth=0;
	lf.lfEscapement=0;
	lf.lfOrientation=0;
	lf.lfWeight=stroke==1?FW_BOLD:FW_NORMAL;
	lf.lfItalic=style;            //??
	lf.lfUnderline=dd->ActiveFormat.ActiveFont.ul;
	lf.lfStrikeOut=dd->ActiveFormat.ActiveFont.strike;
	lf.lfCharSet=ANSI_CHARSET;
	lf.lfOutPrecision=OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision=CLIP_DEFAULT_PRECIS;
	lf.lfQuality=PROOF_QUALITY;
	lf.lfPitchAndFamily=(unsigned char)((spacing==FIX_SPACING?FIXED_PITCH:VARIABLE_PITCH)|family);
	strncpy(lf.lfFaceName,"",32);
	fnt=CreateFontIndirect(&lf);
	if(dd->OrgFnt==NULL)
		dd->OrgFnt=SelectObject(dd->ghdc,fnt);
	else
		DeleteObject(SelectObject(dd->ghdc,fnt));
	
	GetTextMetrics(dd->ghdc,&tm);
	dd->ActiveFormat.ActiveFont.baseline2top=tm.tmAscent*dd->heightpercent/100;
	dd->ActiveFormat.ActiveFont.bottom2baseline=tm.tmDescent*dd->heightpercent/100;
	
}

static void SetTypface(t_driver * dd, t_printerq * q)
{
	
	dd->ActiveFormat.ActiveFont.face = GetParam(q);
	dd->ActiveFormat.ActiveFont.stroke = (unsigned char) (GetParam(q) & 1);
	dd->ActiveFormat.ActiveFont.style = (unsigned char)(GetParam(q) & 1);
	if ((unsigned) dd->ActiveFormat.ActiveFont.face >= 11)
		dd->ActiveFormat.ActiveFont.face = 2;
	print_font(dd);
}

static void IndexUpFromCurrentPos(t_driver *dd,char *s)
{
	int oldYpos=dd->ypos;
	char *out_old,*old_len;
	t_font oldfont=dd->ActiveFormat.ActiveFont;
	dd->ypos-=(int)(dd->rowheight/(dd->heightpercent/100.0)/2.0+0.5);
	dd->ActiveFormat.ActiveFont.size=dd->ActiveFormat.ActiveFont.size*2/3;
	print_font(dd);
	
	out_old=dd->outText;
	old_len=dd->NextToken;
	dd->outText=s;
	dd->NextToken=s+strlen(s);
	printStr(dd);

	dd->ypos=oldYpos;
	dd->ActiveFormat.ActiveFont=oldfont;
	print_font(dd);
	dd->outText=out_old;
	dd->NextToken=old_len;
}

static void PrintOneBar(t_driver *dd,int wide,int bar)
{
	int k;
	int pwidth;
	pwidth=wide?WBARWIDTH(dd->barw):dd->barw;
	for(k=0;k<pwidth;k++)
	{
		MoveToEx(dd->ghdc,dd->xpos,dd->ypos,NULL);
		if(bar)
			LineTo(dd->ghdc,dd->xpos,dd->ypos-(long)(dd->barh*dd->yscaling+0.5));
		dd->xpos++;
	}
	
}

static void PrintBarCodeAtCurPos(t_driver * dd, char *s)
{
/* Printing barcodes of type INTERLEAVED 2/5 with HP-Lasejet-III commands
Interleaved 2/5 prints decimal digits 0-9. The number of digits has to
be even. The digits are printed as pairs, the first one in a pair
constitutes the bars, and the second one constitutes the spaces between
the bars in the former one. The code is self-correcting with 1.4 bit.
The code has different start- and stopsequences. The total width of the
code is not fixed, for a specific digit sequence the total code width
	depends on the number of digits to print. */
   static int bc[10][5] = {
      {0, 0, 1, 1, 0},
      {1, 0, 0, 0, 1},
      {0, 1, 0, 0, 1},
      {1, 1, 0, 0, 0},
      {0, 0, 1, 0, 1},
      {1, 0, 1, 0, 0},
      {0, 1, 1, 0, 0},
      {0, 0, 0, 1, 1},
      {1, 0, 0, 1, 0},
      {0, 1, 0, 1, 0}
   };
	
	/* 1 dot = 1/300 inch = 0.084 mm For HP Lj III: One bar character, '\263',
	advances the print position 83 dots and the width of the visable part
	of the bar is 5 dots. The narrow bar consists of a single bar
	character. Consequently the nbar sholud be followed by backspacing 83-5
	= 78 dots. This displacement in print position is released when
	printing the following space. Analogous the wide bars' (consisting of
	three single bar characters) packspacing of 78 dots after printing the
	last of the three bars is released by the space. As the nbar is 5 dots
	the nspace should be 5 dots wich gives -78+5 = -73 dots in forward
	direction. The single bars in the wbar are overlapping 1 dot -> -83+4??
	= -79 dots. This makes the wbar 13 dots wide wich makes it a factor 2.6
	wider than the narrow one. The acceptable range for the factor is 2.2 -
	3.0 for modulwidth less than 0.5mm, and our printed code is 5*0.084 =
	0.42mm. The wspace is calculated: -78+13 = -65 dots. Corresponding
	values for Canon is: 5-6   13-13 this makes 1:2.17  This is acceptable
	for a moule-width of 6*0.084 = 0.504 > 0.5   The character width is 100
	dots so; the offset from bars is 100-6 = 94 backwards nspace = -94+6
	forward wspace = -94+13 forward wbar =
	modul+backspace(100-4)+modul+backspace(100-3)+modul */
	int i, j, n;
	
	n = strlen(s);
	dprintf("Printing bar-code: %s @(%d,%d)\n",s,dd->xpos,dd->ypos);
	
	/* The startsequence of bars: */
	PrintOneBar(dd,0,1);
	PrintOneBar(dd,0,0);
	PrintOneBar(dd,0,1);
	PrintOneBar(dd,0,0);
	
	
	for (i = 0; i < n; i += 2) { /* print two digits at a time */
		for (j = 0; j < 5; j++)
		{
			PrintOneBar(dd,bc[s[i]-'0'][j],1);
			PrintOneBar(dd,bc[s[i+1]-'0'][j],0);
		}
	}
	/* the stopsequence of bars: */
	PrintOneBar(dd,1,1);
	PrintOneBar(dd,0,0);
	PrintOneBar(dd,0,1);
}


static struct t_driver *CpriStdWinInit(struct t_printerq *q)
{
	t_driver *dd;
	char ret[256],*spooler=NULL,*printerName;
	int i;
	PRINTDLG pd={0};
	
	dd = GetMem0(t_driver, 1);
	dd->outText=GetMem0(char,MAXROWLENGTH);
	dd->q = q;
	dd->macros=NULL;
	dd->ActiveFormat.ActiveFont.size = 12;
	dd->ActiveFormat.ActiveFont.face = COURIER;
	
	pd.lStructSize=sizeof(PRINTDLG);
	pd.Flags=PD_RETURNDC|PD_RETURNDEFAULT;
	
	PrintDlg(&pd);

	dd->ghdc=pd.hDC;
	dd->hdvmd=pd.hDevMode;
	dd->dvmd=NULL;
//	free(pd.hDevNames);
/*
	GetProfileString("windows","device","a",ret,256);   //Get default printer
	printerName=ret;

	for(i=0;ret[i]!='\0';i++)
	{
		if(ret[i]==',')
		{
			ret[i]='\0';
			if(spooler==NULL)
				spooler=ret+i+1;
		}
	}
#if USE_STD_PRINTER != 1
	// printerName="\\\\Terrtom\\EPSON";      //Temporary just for testing purposes
	printerName="Acrobat PDFWriter";    //Temporary just for testing purposes
#endif
	dd->ghdc = CreateDC (spooler, printerName, NULL, NULL);
*/	
	dd->di.cbSize      = sizeof(DOCINFO);
	dd->di.lpszDocName = "Unknown";
	dd->di.lpszOutput  = NULL;
	return dd;
}

static void CpriStdWinClose(struct t_driver *dd)
{
	if(dd)
	{
		if (dd->ActiveFormat.tf.tabs)
			_CpriDestroyTabFrame(&dd->ActiveFormat.tf);
		if (dd->HeadFormat.tf.tabs)
			_CpriDestroyTabFrame(&dd->HeadFormat.tf);
		Release(dd->outText);
		deleteMacros(&dd->macros);
		deleteMacros(&dd->head.next);
		DeleteDC(dd->ghdc);
		Release(dd);
	}
}

static void ResetDriverState(t_driver * dd)
{
	t_tab *tabs;
	int n, copies, nocopies;
	char *nspace;
	char *wspace;
	char *wbar;
	
	
	nspace = dd->nspace;
	wspace = dd->wspace;
	wbar = dd->wbar;
	n = dd->ActiveFormat.tf.n;
	nocopies = dd->nocopies;
	copies = dd->copies;
	tabs = dd->ActiveFormat.tf.tabs;
	//memset(dd, 0, sizeof(t_driver));
	dd->ActiveFormat.tf.tabs = tabs;
	dd->ActiveFormat.tf.n = n;
	dd->nocopies = nocopies;
	dd->copies = copies;
	dd->nspace = nspace;
	dd->wspace = wspace;
	dd->wbar = wbar;
	
	
	GetRes(dd);
	dd->bottom = dd->pagelen - 150;
	dd->heightpercent = 100;
	dd->nocopies = 1;
	dd->ActiveFormat.ActiveFont.size = 12;
	dd->ActiveFormat.ActiveFont.face = COURIER;
	deleteMacros(&dd->macros);
	dd->xpos=0;
	dd->ypos=0;
}

static void driver(t_printerq * q)
{
	char s[200];
	char c;
	int param, barw, barh, pattern, grey, embedded,uscp;
	double fwidth, fheight;
	t_driver *dd;
	
	dd=q->drv;
	dd->NextToken=dd->outText;
	*dd->outText='\0';

	while(dd->pos<dd->bufsize)
	{
		c=dd->prnbuf[dd->pos++];
		switch (c) {
		case 27:
			printStr(dd);
			c=dd->prnbuf[dd->pos++];
			switch (c) {
			case EPC_ABS_MOVEY:
				param = (int)((uscp=GetParam(q)) * dd->yscaling + 0.5);
				
				dd->ypos = param-dd->yoffset;
				dprintf("ABSMOVEY to %d\n",dd->ypos);
				dprintf("R_ABSMOVEY %d\n",uscp);
				break;
			case EPC_REL_MOVEX:
				param = (int)((uscp=GetParam(q)) * dd->xscaling + 0.5);
				dd->xpos+=param;
				dprintf("R_RELMOVEX %d\n",uscp);
				break;
			case EPC_REL_MOVEY:
				param = (int)((uscp=GetParam(q)) * dd->yscaling + 0.5);
				dd->ypos += param;
				dprintf("R_RELMOVEY %d\n",uscp);
				break;
			case EPC_ABS_MOVEX:
				param = (int)((uscp=GetParam(q)) * dd->xscaling + 0.5);
				dd->xpos=param-dd->xoffset;
				dprintf("ABSMOVEX to %d\n",dd->xpos);
				dprintf("R_ABSMOVEX %d\n",uscp);
				break;
			case EPC_RECTANGLE:
				fwidth = GetFloatParam(q) * dd->xscaling;
				fheight = GetFloatParam(q) * dd->yscaling;
				dprintf("R_RECT %f %f\n",fwidth/dd->xscaling,fheight/dd->yscaling);
				
				DrawRect(dd->ghdc,dd->xpos,dd->ypos,dd->xpos+fwidth,dd->ypos+fheight);
				
				dd->xpos+=(long)fwidth;
				
				break;
			case EPC_MACROID:
				param = GetParam(q);
				dd->CurrentMacro=param;
				
				break;
			case EPC_NR_OF_COPIES:
				if (!dd->copies)
					dd->nocopies = GetParam(q);
				else
					GetParam(q);
				
				break;
			case EPC_BEGIN_TO_BE_REPEATED:
				dd->HeadFormat=dd->ActiveFormat;
				_CpriCopyTabFrame(&dd->ActiveFormat.tf,&dd->HeadFormat.tf);
				doRecordMacro(dd, &dd->head);
				PlayHeadMacro(q);
				break;
			case EPC_END_TO_BE_REPEATED:
				
				break;
			case EPC_START_RECORD_MACRO:
				RecordMacro(dd,dd->CurrentMacro,&dd->macros);
				break;
			case EPC_EXECUTE_MACRO:
				PlayMacro(q,dd->CurrentMacro,dd->macros);
				break;
				
			case EPC_TABLE:
				dd->tablemode= 1;
				//x
				param=GetParam(q);
				_CpriInitTabs(&dd->ActiveFormat.tf,0,dd->xscaling);
				if(param*dd->xscaling<dd->xoffset)
					param=dd->xoffset/dd->xscaling+1;
				_CpriSetSoftMargin(&dd->ActiveFormat.tf, &dd->ActiveFormat.margin, param, dd->xscaling);
				//y
				dd->topmargin = (int)((GetParam(q)) * dd->yscaling + 0.5)-dd->yoffset;
				if(dd->ypos<dd->topmargin)
					dd->ypos=dd->topmargin;
				dd->linebreaking=1;
				//bottom
				dd->bottom = (int)(GetParam(q) * dd->yscaling + 0.5);
				if(dd->bottom<dd->minbottom)
					dd->bottom=dd->minbottom;
				dd->bottom+=dd->yoffset;
				break;
			case EPC_END_TABLE:
				dd->tablemode=0;
				break;
			case EPC_TEXT_TO_HLINE_SPACE:
				GetParam(q);
				break;
			case EPC_HORIZONTAL_BAR:
				barw = GetParam(q);
				barh = GetParam(q);
				pattern = GetParam(q);
				grey = GetParam(q);
				embedded = GetParam(q);
				//          MakeBar(q, dd, barw, barh, pattern, grey, embedded);
				
				break;
			case EPC_END_BAR_LABEL:
//				dd->printing_embedded = 0;
				
				break;
			case EPC_CREATE_COLUMN:          /* settab */
				_CpriSetTab(&dd->ActiveFormat.tf, GetParam(q), dd->xscaling);
				
				break;
			case EPC_ACTIVATE_COLUMNS:
				_CpriActivateTabs(&dd->ActiveFormat.tf);
				
				break;
			case EPC_DEACTIVATE_COLUMN:
				_CpriRemovePriTab(&dd->ActiveFormat.tf, GetParam(q));
				
				break;
			case EPC_MERGE_ALL_COLUMNS:
				_CpriMergeAllColumns(&dd->ActiveFormat.tf);
				break;
			case EPC_MOVE_TO_NEXT_COLUMN:       /* tab to next */
				if(dd->linebreaking)
					FinnishLineBreak(dd);
				if (dd->ActiveFormat.tf.curtab >= dd->ActiveFormat.tf.n)
					break;
				dd->xpos=dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled-dd->xoffset+(int)(dd->xscaling+0.5);
				dprintf("Jumping to tab %d at pos %d = %d mm\n",dd->ActiveFormat.tf.curtab,dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled,dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].mm);
				_CpriMoveToNextCell(&dd->ActiveFormat.tf);
				break;
			case EPC_RESET:           /* clear all stuff, send printable area,
									  default symbols(windows) for primary set
									  and default symbols(PC8) and fix-spacing
				for secondary set */
				//          ResetDriverState(dd);
				//          _CpriSetSoftMargin(tf, &dd->margin, 5, dd->scaling, 5);
				//          fprintf(stderr,"Resetting at pos %d\n",dd->pos-2);
				break;
			case EPC_REMOVE_COLUMNS:       /* clear tabs */
				ClearTabs(&dd->ActiveFormat.tf);
				
				break;
			case EPC_START_GRID:   
				dd->ActiveFormat.gridon = 1;
				break;
			case EPC_END_GRID:   
				dd->ActiveFormat.gridon = 0;
				break;
			case EPC_SCALING:
				param = GetParam(q);
				break;
			case EPC_LINE_HEIGHT:
				dd->heightpercent = GetParam(q);  /* given in % of "normal height"
				i.e. 4.2 mm/12dpi */
				break;
			case EPC_NEW_LINE:
				NewLine(dd);
				break;
			case EPC_NEW_PAGE:
				NewPage(dd);
				break;
			case EPC_DOSTEXT:
				GetParam(q);
				break;
			case EPC_ALIGN:
				param = GetParam(q);
				GetSubsequentText(dd, s);
				strcpy(dd->NextToken,s);
				dd->NextToken+=strlen(s);
				switch (param) {
					int width,i,xpos;
				case ALIGN_LEFT:
					SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
					break;
				case ALIGN_CENTER:
					SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_CENTER|TA_UPDATECP);
					dd->xpos=(dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled+dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.prevtab].scaled)/2-dd->xoffset;
					break;
				case ALIGN_RIGHTHE:
					{
						SIZE size;
						if(!GetTextExtentPoint32(dd->ghdc,"",1,&size))
						{
							DWORD err;
							err=GetLastError();
						}
						width=size.cx;
						xpos=dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled-dd->xoffset-width-(int)(dd->xscaling+0.5);
						dd->xpos=xpos;
						if(dd->NextToken[-1]=='')
							dd->NextToken-=1;
						SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_RIGHT|TA_UPDATECP);
						printStr(dd);
						dd->xpos=xpos;
						SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
						if(s[strlen(s)-1]=='')
						{
							*dd->NextToken='';
							*dd->NextToken++;
						}
					}
					break;

				case ALIGN_RIGHTDP:
					if(!GetCharWidth32(dd->ghdc,'5','5',&width))
					{
						DWORD err;
						err=GetLastError();
					}
					xpos=dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled-dd->xoffset-width-(int)(dd->xscaling+0.5);
					if(!GetCharWidth32(dd->ghdc,'.','.',&width))
					{
						DWORD err;
						err=GetLastError();
					}
					xpos-=width;
					dd->xpos=xpos;
					for(dd->NextToken=dd->outText;*dd->NextToken!='\0'&&*dd->NextToken!='.'&&*dd->NextToken!=',';dd->NextToken++)
						;
					SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_RIGHT|TA_UPDATECP);
					printStr(dd);
					dd->xpos=xpos;
					for(i=0;s[i]!='\0'&&s[i]!='.'&&s[i]!=',';i++)
						;
					SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
					strcpy(dd->outText,s+i);
					dd->NextToken+=strlen(dd->NextToken);
					break;
				case ALIGN_RIGHT:
					SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_RIGHT|TA_UPDATECP);
					dd->xpos=dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled-dd->xoffset-(int)(dd->xscaling+0.5);
					break;
				}
				printStr(dd);
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
				break;
			case EPC_SUP_RALIGN_CELL:
				dd->xpos=dd->ActiveFormat.tf.tabs[dd->ActiveFormat.tf.curtab].scaled-dd->xoffset-(int)(dd->xscaling+0.5);
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_RIGHT|TA_UPDATECP);
			case EPC_SUP:
				GetSubsequentText(dd, s);
				IndexUpFromCurrentPos(dd, s);
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
				break;
			case EPC_SUP_LALIGN:
				GetSubsequentText(dd, s);
				dd->IndexWaiting=SUP_TEXT_TO_LEFT;
				dd->IndexedText=MkString(s);
				break;
			case EPC_SUP_RALIGN:
				GetSubsequentText(dd, s);
				dd->IndexWaiting=SUP_TEXT_TO_RIGHT;
				dd->IndexedText=MkString(s);
				break;
			case EPC_STRIKE:
				dd->ActiveFormat.ActiveFont.strike=1;
				print_font(dd);
				GetSubsequentText(dd, s);
				strcpy(dd->NextToken,s);
				dd->NextToken+=strlen(s);
				printStr(dd);
				dd->ActiveFormat.ActiveFont.strike=0;
				print_font(dd);
				//          MakeStrikeText(dd, s);
				break;
			case EPC_HTML_TAG:      /* html-tag - clear input and ignore */
				param = GetParam(q);
				
				break;
			case EPC_END_HTML_TAG:   /* html-tagend  - ignore */
				param=GetParam(q);
				
				break;
			case EPC_BAR_CODE:
				GetSubsequentText(dd, s);
				PrintBarCodeAtCurPos(dd, s);
				
				break;
			case EPC_BAR_CODE_W:
				dd->barw = (int)(GetParam(q)/100.0*dd->xscaling+0.5);
				break;
			case EPC_BAR_CODE_H:
				dd->barh = GetParam(q);
				break;
			case EPC_CHECK:
				{
					POINT p[5]={{0,-4},{1,-4},{2,0},{4,-8},{5,-8}};
					int i;
					for(i=0;i<5;i++)
					{
						p[i].x=p[i].x*dd->xscaling/2/12*dd->ActiveFormat.ActiveFont.size+dd->xpos+0.5;
						p[i].y=p[i].y*dd->yscaling/2/12*dd->ActiveFormat.ActiveFont.size+dd->ypos+0.5;
					}
					
					Polyline(dd->ghdc,p,5);
					dd->xpos=p[4].x;
				}
				break;

			case EPC_FONT_TYPEFACE:
				SetTypface(dd,  q);
				
				break;
			case EPC_FONT_SIZE:
				dd->ActiveFormat.ActiveFont.size = GetParam(q);
				print_font(dd);
				//          ModifyLineHeight(dd);
				
				break;
			case EPC_UNDERLINE:
				dd->ActiveFormat.ActiveFont.ul = GetParam(q);
				
				break;
			case EPC_PORTRAIT:
			case EPC_LANDSCAPE:
				dd->dvmd->dmFields|=DM_ORIENTATION;
				if ( c== EPC_PORTRAIT)
					dd->dvmd->dmOrientation=DMORIENT_PORTRAIT;
				else
					dd->dvmd->dmOrientation=DMORIENT_LANDSCAPE;
				break;
			case EPC_END_START_SEQUENCE:
				if(!ResetDC(dd->ghdc,dd->dvmd))
				{
					DWORD error;
					error=GetLastError();
#ifdef DEBUG
					assert(0);
#endif
				}
				
				StartPage(dd->ghdc);
				SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
				GetRes(dd);
				if(dd->barw==0)
					dd->barw=(int)(NBARWIDTH*dd->xscaling+0.5);
				break;
			case EPC_CASSETTE_FEED:
			case EPC_MANUAL_FEED:
				dd->dvmd->dmFields|=DM_DEFAULTSOURCE;
				if(c==EPC_CASSETTE_FEED)
					dd->dvmd->dmDefaultSource=DMBIN_AUTO;
				else
					dd->dvmd->dmDefaultSource=DMBIN_MANUAL;
				break;
			case EPC_TEXT_TO_VLINE_SPACE: //????????????
			case EPC_FIX_ROW_HEIGHT:     //????????????
			GetParam(q);
			break;
				
			case EPC_PUSH_POS:
				*dd->posStackPos++=dd->xpos;
				*dd->posStackPos++=dd->ypos;
				dprintf("R_push\n");
				
				break;
			case EPC_POP_POS:
				dd->ypos=*--dd->posStackPos;
				dd->xpos=*--dd->posStackPos;
				dprintf("R_pop\n");
				break;
				
			case EPC_DELETE_MACROS:
				deleteMacros(&dd->macros);
				break;
				
				
			default:
				dprintf("Unknown EPC code %d @ %d\n",c,dd->pos-1);
				break;
         }
         break;
      case '\n':
		  NewLine(dd);
		  break;
      default:
		  *dd->NextToken=c;
		  dd->NextToken++;
		  
		  break;
      }
   }
   printStr(dd);
}

static int CpriStdWinDriver(t_driver *dd)
{
	int bytesRead=0;
	t_printerq *q;

	q = dd->q;
	dd->dvmd=GlobalLock(dd->hdvmd);

	_CpriInitBufferedInput(q);
	dd->prnbuf = _CpriGetInputPos(q);
	
	dd->di.lpszDocName= _CpriProcessedName(q);
	do
	{
		StartDoc(q->drv->ghdc,&q->drv->di);
		SetTextAlign(dd->ghdc,VERT_TEXT_ALIGN|TA_LEFT|TA_UPDATECP);
		GetRes(q->drv);
		dd->CurrentMacro=-1;
		dd->posStackPos=dd->posStack;
		dd->pos=0;
		dd->xpos=0;
		dd->ypos=0;
		dd->tablemode=0;
		dd->linebreaking=1;
		dd->ActiveFormat.margin.mm=0;
		dd->ActiveFormat.margin.scaled=0;
		dd->barh=BARCODEHEIGHT;
		dd->barw=0;

		q->drv->bufsize=_CpriGetInputSize(q)-2; /* skip the 2 last control chars */
		q->drv->dvmd->dmFields=0;
		
		dd->OrgFnt=NULL;
		print_font(dd);
		
		driver(q);
		if (--q->drv->nocopies > 0)
		{
			q->drv->copies = 1;
		}
		else
		{
			q->drv->copies = 0;
		}
		DeleteObject(SelectObject(dd->ghdc,dd->OrgFnt));
		dd->OrgFnt=NULL;
		EndPage(q->drv->ghdc);
		EndDoc(q->drv->ghdc);
	}while(q->drv->copies>0);

   _CpriPopFront(q);
	GlobalUnlock(dd->hdvmd);
	dd->dvmd=NULL;
//	Release((void*)dd->di.lpszDocName);
	return EOF;
	
}

static t_cpri_vtable vtable = {CpriStdWinInit, CpriStdWinClose, CpriStdWinDriver};
t_cpri_vtable *_cpri_stdwin_vtable = &vtable;
#endif
