/* Module DRV_TXT.C
   Contains the print-driver for creating text-files */

#include <string.h>
#include <allegro.h>

#include <cgui.h>
#include <cgui/mem.h>

#include "cpri.h"
#include "spooler.h"
#include "pridrive.h"
#include "drv_txt.h"
#include "drv_common.h"
#include "dialpri.h"
#include "cpri.ht"
#include "labldata.h"

#define QBUFSIZ 500
#define AddCr(s, cr) *s++ = '\n';if (cr) *s++ = cr
#define DST_PRINTER 0
#define DST_FILE  1
typedef struct t_driver {
   t_printerq *q;
   t_tab_frame tf;              /* A list of the current tabs (in three
                                   different scalings) */
   t_tab margin;                /* Specifies the current margins. */
   int topmargin;
   int ypos;                    /* current printing y-position */
   int r2line;                  /* vertical distance from text-row to the
                                   horizontal line in a table */
   int lh;                      /* line height */
   int bottom;                  /* "soft" bottom margin */
   int gridon;                  /* It is necessary to set gridon after
                                   defining the set of tabs to be used */
   int indexwidth;
   int dos;
   char *buf;                   /* buffer containg one "row" of text (one row
                                   might contain several rows if grids are
                                   used also additional linefeeds if extended
                                   lineheight) */
   char *p;                     /* pointer to curren position in buf. */
   int n;
   int size;
   int eof;
} t_driver;


static char *DrawFrame(t_driver * dd, char left, char mid, char right,
                       char fill, t_tab_frame * tf)
{
   int i, j;
   t_tab *tabs;
   char *p;

   p = dd->p - dd->n;
   tabs = tf->tabs;
   for (i = 0; i < tabs->block - 1; i++)
      *p++ = ' ';
   *p++ = left;
   for (j = 1, tabs++; j < tf->n; j++, tabs++) {   /* drop first and last */
      for (i++; i < tabs->block - 1; i++)
         *p++ = fill;
      if (j < tf->n - 1)
         *p++ = mid;
   }
   *p++ = right;
   *p = 0;
   dd->n += p - dd->p;
   return p;
}

static void TabToNext(t_driver * dd, t_tab_frame * tf)
{
   int tab;
   char *p;

   if (tf->curtab >= tf->n)
      return;
   tab = tf->tabs[tf->curtab].block;
   if (dd->gridon)
      tab--;
   p = dd->p;
   if (tab <= dd->n) {
      if (p > dd->buf && dd->n > 0) {
         p -= (dd->n - tab);
         dd->n = tab;
      } else {
      }
   } else {
      for (; dd->n < tab; dd->n++)
         *p++ = ' ';
   }
   if (dd->gridon) {
      *p++ = '';
      dd->n++;
   }
   tf->curtab++;
   while (tf->tabs[tf->curtab].inactive && tf->curtab < tf->n) {
      if (dd->gridon) {
         tab = tf->tabs[tf->curtab].block - 1;
         for (; dd->n < tab; dd->n++)
            *p++ = ' ';
         *p++ = '';
         dd->n++;
      }
      tf->curtab++;
   }
   dd->p = p;
}

static void RightAlignText(t_driver * dd, t_tab_frame * tf, char *s, int c)
{
   int w;
   char fmt[100], *p;

   p = dd->p;
   if (tf->curtab < tf->n && tf->curtab > 0) {
      w = tf->tabs[tf->curtab].block - tf->tabs[tf->curtab - 1].block - 1;
      if (c == ALIGN_RIGHTHE)
         if ((dd->dos && s[strlen(s) - 1] != '')
             || (!dd->dos && s[strlen(s) - 1] != HALF_CHAR))
            w--;
      if (c == ALIGN_RIGHTDP) {
         if (s[strlen(s) - 1] == '.')
            w--;
         else if (s[strlen(s) - 2] == '.');
         else
            w -= 2;
      }
      sprintf(fmt, "%c%ds", '%', w);
      p += sprintf(p, fmt, s);
      dd->n += w;
   } else {
      p += sprintf(p, s);
      dd->n += strlen(s);
   }
   dd->p = p;
}

static void CenterText(t_driver * dd, t_tab_frame * tf, char *s)
{
   int i, w, sp;
   char *p;

   p = dd->p;
   if (tf->curtab < tf->n && tf->curtab > 0) {
      w = tf->tabs[tf->curtab].block - tf->tabs[tf->curtab - 1].block;
      if (dd->gridon)
         w--;
      sp = (w - (int) strlen(s)) / 2;
      for (i = 0; i < sp; i++)
         *p++ = ' ';
      p += sprintf(p, s);
   } else {
      p += sprintf(p, s);
   }
   dd->n += p - dd->p;
   dd->p = p;
}

extern t_driver *TextInit(t_printerq *q)
{
   t_driver *dd;

   dd = GetMem0(t_driver, 1);
   dd->buf = GetMem0(char, 1000);
   dd->p = dd->buf;
   dd->size = 1000;
   dd->q = q;
   return dd;
}

extern void TextClose(t_driver *dd)
{
   _CpriDestroyTabFrame(&dd->tf);
   if (dd->buf)
      Release(dd->buf);
   Release(dd);
}

static void MakeTextRow(t_driver *dd, char cr,int dest)
{
   char s[500], fmt[100], *p;
   int i, j, c, x, n, len, param, oldmargin;
   t_tab_frame *tf;
   double ddummy;
   t_printerq *q;

   tf = &dd->tf;
   q = dd->q;
   dd->p = dd->buf;
   while (1) {                  /* a line-feed command will break */
      c = _CpriGetChar(q);
      switch (c) {
      case 27:
         c = _CpriGetChar(q);
         switch (c) {
         case EPC_ACTIVATE_COLUMNS:
            _CpriActivateTabs(tf);
            break;
         case EPC_DEACTIVATE_COLUMN:
            _CpriGetIntParam(q, &param);
            _CpriRemovePriTab(tf, param);
            break;
         case EPC_TEXT_TO_HLINE_SPACE:
            _CpriGetIntParam(q, &param);
            dd->r2line = ((float) param * dd->lh) / 100 + 0.5;
            break;
         case EPC_LINE_HEIGHT:
            _CpriGetIntParam(q, &param);
            dd->lh = (param + 50) / 100;
            if (dd->gridon && dd->lh < 2)
               dd->lh = 2;
            if (dd->gridon && dd->r2line)
               dd->r2line = ((float) dd->r2line / dd->lh) + 0.5;
            break;
         case EPC_DOSTEXT:
            _CpriGetIntParam(q, &dd->dos);
            break;
         case EPC_SUP:
         case EPC_SUP_RALIGN_CELL:
            p = dd->p;
            *p++ = '~';
            _CpriGetStringParam(q, p);
            p += strlen(p);
            dd->n += p - dd->p;
            dd->p = p;
            break;
         case EPC_SUP_RALIGN:
            p = dd->p;
            _CpriGetStringParam(q, s);
            p += strlen(p);
            sprintf(fmt, "%c%ds", '%', dd->indexwidth);
            p += sprintf(p, fmt, s);
            if (*s)
               *p++ = '~';
            else
               *p++ = ' ';
            dd->n += p - dd->p;
            dd->p = p;
            break;
         case EPC_CHECK:
            *dd->p++ = '\373';
            dd->n++;
            break;
         case EPC_ALIGN:
            _CpriGetIntParam(q, &param);
            _CpriGetStringParam(q, s);
            switch (param) {
            case ALIGN_LEFT:
               strcpy(dd->p, s);
               len = strlen(s);
               dd->p += len;
               dd->n += len;
               break;
            case ALIGN_CENTER:
               CenterText(dd, tf, s);
               break;
            case ALIGN_RIGHT:
            case ALIGN_RIGHTHE:
            case ALIGN_RIGHTDP:
               RightAlignText(dd, tf, s, param);
               break;
            }
            break;
         case EPC_ABS_MOVEX:
            _CpriGetIntParam(q, &x);
            n = (x / 2.54 + 0.5) - dd->margin.block;
            for (; dd->n < n; dd->n++)
               *dd->p++ = ' ';
            break;
         case EPC_CREATE_COLUMN:
            _CpriGetIntParam(q, &param);
            _CpriSetTab(&dd->tf, param, 0);
            if (tf->tabs[0].block == 0 && dd->gridon) {
               tf->tabs[0].block++;
               tf->tabs[1].block++;
            }
            break;
         case EPC_TABLE:
            oldmargin = dd->margin.block;
            _CpriGetIntParam(q, &param);
            _CpriInitTabs(tf, param, 0);
            if (oldmargin < dd->margin.block) {
               p = dd->p;
               for (i = dd->n; i < dd->margin.block; i++)
                  *p++ = ' ';
               dd->n = i;
               dd->p = p;
            } else if (dd->n - oldmargin > 0) {
               p = dd->p - dd->n;
               memmove(p + dd->margin.block, p + oldmargin,
                       dd->n - oldmargin);
               dd->p -= oldmargin - dd->margin.block;
               dd->n -= oldmargin - dd->margin.block;
            }
            _CpriGetIntParam(q, &param);
            dd->topmargin = param / 4.33 + 0.5 - 1;
            /* 4.33 mm default "hard" margin */
            if (dd->topmargin < 0)
               dd->topmargin = 0;
            for (i = dd->ypos; i < dd->topmargin; i++) {
               AddCr(dd->p, cr);
            }
            if (i > dd->ypos) {
               dd->ypos = i;
               for (i = 0; i < dd->n; i++)
                  *dd->p++ = ' ';
            }
            _CpriGetIntParam(q, &param);
            dd->bottom = 66 - param / 4.33 + 0.5;/* 4.33 mm default "hard" margin */
            break;
         case EPC_NEW_LINE:
            if (tf->curtab)
               TabToNext(dd, tf);
            dd->n = 0;
            for (i = 0; i < dd->r2line; i++) {
               if (i > 0 && dd->gridon) {
                  dd->p = DrawFrame(dd, '', '', '', ' ', tf);
                  dd->n = 0;
               }
               AddCr(dd->p, cr);
            }
            if (dd->gridon) {
               dd->p = DrawFrame(dd, '', '', '', '', tf);
               dd->n = 0;
               AddCr(dd->p, cr);
               i++;
            }
            for (j = 0; i < dd->lh; i++, j++) {
               if (dd->gridon) {
                  dd->p = DrawFrame(dd, '', '', '', ' ', tf);
                  dd->n = 0;
               }
               AddCr(dd->p, cr);
            }
            if (++dd->ypos + 1 > dd->bottom&&dest==DST_PRINTER) {
               *dd->p++ = '\14';
               *dd->p++ = '\15';
               for (i = 0; i < dd->topmargin; i++) {
                  AddCr(dd->p, cr);
               }
               dd->ypos = i;
            }
            dd->n = 0;
            tf->curtab = 0;
            *dd->p = 0;
            dd->p = dd->buf;
            return;
            break;
         case EPC_NEW_PAGE:
            if(dest==DST_PRINTER)
            *dd->p++ = '\14';
            if (cr)
               *dd->p++ = cr;
            else
               *dd->p++ = '\n';
            *dd->p = 0;
            dd->n = 0;
            dd->p = dd->buf;
            return;
            break;
         case EPC_RESET:
            dd->dos = 0;
            dd->topmargin = dd->ypos = 0;
            dd->bottom = 62;
            tf->n = 0;
            dd->gridon = 0;
            break;
         case EPC_REMOVE_COLUMNS:
            tf->n = 0;
            dd->gridon = 0;
            break;
         case EPC_MERGE_ALL_COLUMNS:
            _CpriMergeAllColumns(tf); /* This is probably not enough..*/
            break;
         case EPC_END_GRID:
            dd->gridon = 0;
            dd->n = 0;
            p = DrawFrame(dd, '', '', '', '', tf);
            AddCr(p, cr);
            for (i = 0; i < dd->margin.block; i++)
               *p++ = ' ';
            dd->n = i;
            *p = 0;
            dd->p = dd->buf;
            return;
            break;
         case EPC_START_GRID:
            if (dd->gridon && dd->lh < 2)
               dd->lh = 2;
            if (dd->r2line)
               dd->r2line = ((float) dd->r2line / dd->lh) + 0.5;
            else
               dd->r2line = 1;
            if (!dd->gridon) {
               if (tf->n && tf->tabs[0].block == 0) {
                  for (i = 0; i < tf->n; i++)
                     tf->tabs[i].block++;
               }
               dd->gridon = 1;
               tf->curtab = 0;
               p = dd->p;
               AddCr(p, cr);
               p = DrawFrame(dd, '', '', '', '', tf);
               AddCr(p, cr);
               for (i = 0; i < dd->margin.block; i++)
                  *p++ = ' ';
               dd->n = i;
               dd->p = p;
            }
            break;
         case EPC_STRIKE:       /* not possible to implement */
            _CpriGetStringParam(q, dd->p);
            dd->p += strlen(dd->p);
            *dd->p++ = '';
            *dd->p = 0;
            break;
         case EPC_MOVE_TO_NEXT_COLUMN:    /* tab to next */
            TabToNext(dd, tf);
            break;
         case EPC_HORIZONTAL_BAR:     /* unusued 5-param-command */
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            break;
         case EPC_FONT_TYPEFACE:
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            _CpriGetIntParam(q, &param);
            break;
         case EPC_REL_MOVEX:
            _CpriGetFloatParam(q, &ddummy);
            break;
         case EPC_RECTANGLE:
            _CpriGetFloatParam(q, &ddummy);
            _CpriGetFloatParam(q, &ddummy);
            break;
         case EPC_BAR_CODE:
         case EPC_HTML_TAG:
            _CpriGetStringParam(q, s);
            break;
         case EPC_ABS_MOVEY:
         case EPC_BAR_CODE_W:
         case EPC_BAR_CODE_H:
         case EPC_SCALING:
         case EPC_TEXT_TO_VLINE_SPACE:
         case EPC_FONT_SIZE:
         case EPC_UNDERLINE:
         case EPC_REL_MOVEY:
         case EPC_MACROID:
         case EPC_NR_OF_COPIES:
         case EPC_FIX_ROW_HEIGHT:
            _CpriGetIntParam(q, &param);
            break;
         case EPC_CASSETTE_FEED:
         case EPC_MANUAL_FEED:
         case EPC_END_TABLE:
         case EPC_BEGIN_TO_BE_REPEATED:
         case EPC_END_TO_BE_REPEATED:
         case EPC_PORTRAIT:
         case EPC_LANDSCAPE:
         case EPC_END_HTML_TAG:
         case EPC_END_START_SEQUENCE:
         case EPC_END_BAR_LABEL:
         case EPC_PUSH_POS:
         case EPC_POP_POS:
         case EPC_START_RECORD_MACRO:
         case EPC_STOP_RECORD_MACRO:
         case EPC_EXECUTE_MACRO:
         case EPC_DELETE_MACROS:
            break;
         }
         break;
      case '\n':
         dd->n = 0;
         tf->curtab = 0;
         dd->ypos++;
         AddCr(dd->p, cr);
         break;
      case EOF:
         *dd->p = 0;
         dd->eof = 1;
         dd->p = dd->buf;
         return;
         break;
      default:
         if(c>=32&&dest==DST_FILE)
       {
          if (dd->dos)
            c = _CpriToDOS(c, dd->dos);
          dd->n++;
          *dd->p++ = c;
       }
         break;
      }
   }
}

/* Driver for plain text-printer (unknown printer). It will just return
   next char from the current row until row ends, when it will ask for
   a new row to be built. */
static int TextDriver(t_driver *dd)
{
   if (*dd->p == 0) {
      if (dd->eof) {
         dd->eof = 0;
         return EOF;
      }
      MakeTextRow(dd, '\15',DST_PRINTER);
   }
   return *dd->p++;
}

/* Shell function that makes use of the text-driver for creating a text file */
static int TextFileDriver(t_driver *dd)
{
   FILE *dst;
   char *fn;

   fn = _CpriGetDestinationFilename(dd->q);

   dst = fopen(fn, "w");
   if (dst) {
      do {
         MakeTextRow(dd, 0,DST_FILE);
         fprintf(dst, dd->buf);
      } while (!dd->eof);
      dd->eof=0;
     fclose(dst);
   }
   return EOF;
}

t_cpri_vtable unknown_vtable = {TextInit, TextClose, TextDriver};
t_cpri_vtable *_cpri_unkown_vtable = &unknown_vtable;

t_cpri_vtable textfile_vtable = {TextInit, TextClose, TextFileDriver};
t_cpri_vtable *_cpri_text_vtable = &textfile_vtable;
