/* Module PRINTINI.C
   This file contains functions for initializing the printer system (loading
   data from file and saving to file) */

#include <string.h>
#include <allegro.h>
#include <cgui.h>
#include <cgui/mem.h>

#include "labldata.h"
#include "dialpri.h"
#include "spooler.h"
#include "printini.h"
#include "cpri.h"

static char *print_section_name[] =
    { "Printer:LPT1", "Printer:LPT2", "Printer:LPT3", "Printer:DOS",
       "Printer:NET" };
static char *pri_dbid_charset = "Char-set";
static char *pri_dbid_install = "Installed";
static char *pri_dbid_brand = "Brand";
static char *pri_dbid_netcom1 = "Net-command";
static char *pri_dbid_netcom2 = "Command-cont";

char _cpri_spool_dir[MAXPATH];
static int printer_inited;

extern void _CpriSavePrinterSettings(void)
{
   int i;
   t_printerq *q;

   for (i = 0; i < 5; i++) {
      q = printerqs + i;
      set_config_int(print_section_name[i], pri_dbid_charset, q->dostext);
      set_config_int(print_section_name[i], pri_dbid_install,
                     q->installed);
      set_config_int(print_section_name[i], pri_dbid_brand, q->brand);
      if (i == QUEUE_NETPORT) {
         if (*q->net1 == 0)
            strcpy(q->net1, " ");
         set_config_string(print_section_name[i], pri_dbid_netcom1,
                           q->net1);
         if (*q->net2 == 0)
            strcpy(q->net2, " ");
         set_config_string(print_section_name[i], pri_dbid_netcom2,
                           q->net2);
      }
   }
}

static void LoadPrinterSettings(void)
{
   t_printerq *q;
   int i;

   for (i = 0; i < 5; i++) {
      q = printerqs + i;
      q->dostext = get_config_int(print_section_name[i], pri_dbid_charset, 0);
      q->installed = get_config_int(print_section_name[i], pri_dbid_install, 0);
      q->brand = get_config_int(print_section_name[i], pri_dbid_brand, 0);
      if (i == QUEUE_NETPORT) {
         strncpy(q->net1, get_config_string(print_section_name[i], pri_dbid_netcom1, ""), MAXNETCOMMAND);
         strncpy(q->net2, get_config_string(print_section_name[i], pri_dbid_netcom2, ""), MAXNETCOMMAND);
      }
   }
   printerqs[QUEUE_OPTIMALPORT].installed = 1;
   printerqs[QUEUE_REQUEST].installed = 1;
}

extern void CpriPrintingProperties(int port, int use, int brand, const char *cmd1, const char *cmd2)
{
   t_printerq *q;
   int newbrand, newinst;

   CpriInitPrinter();
   if ((unsigned) port < TOT_NR_OF_QUEUES) {
      q = printerqs + port;
      newbrand = q->brand != brand;
      newinst = q->installed && !use;
      if (newinst || newbrand)
         if (q->vtable)
            q->vtable->Close(q->drv);
         q->drv = NULL;
      q->installed = use;
      q->brand = brand;
      strncpy(q->net1, cmd1, MAXNETCOMMAND);
      strncpy(q->net2, cmd2, MAXNETCOMMAND);
      if (q->installed && q->drv == NULL) {
         _CpriSetPrinterDriver(q);
         if (q->vtable)
            q->drv = q->vtable->Init(q);
      }
   }
}

extern int CpriIsSpooling(void)
{
   int i;

   for (i = 0; i < TOT_NR_OF_QUEUES; i++)
      if (!_CpriEmptyQ(printerqs + i))
         return 1;
   return 0;
}

static void DeInitPrinter(void *data nouse)
{
   int i;
   t_printerq *q;

   if (!printer_inited)
      return;
   for (i = 0; i < TOT_NR_OF_QUEUES; i++) {
      q = printerqs + i;
      if (q->installed && q->vtable)
         q->vtable->Close(q->drv);
      q->drv = NULL;
   }
   _CpriDestroyBarIcon();
}

extern void CpriDeInitPrinter(void)
{
   DeInitPrinter(NULL);
}

extern void CpriInitPrinter(void)
{
   int i;
   t_printerq *q;
   char *tempdir;

   if (printer_inited)
      return;
   _CpriInitBarIcons();
   printer_inited = 1;
   if (*_cpri_spool_dir == 0) {
      tempdir = getenv("TEMP");
      if (tempdir == NULL)
         tempdir = getenv("TMP");
      if (tempdir && strlen(tempdir) < 255)
         strncpy(_cpri_spool_dir, tempdir, 255);
   }
   LoadPrinterSettings();
   for (i = 0; i < NR_OF_QUEUES; i++) {
      q = printerqs + i;
      q->qno = i;
      if (q->installed)
         _CpriSetPrinterDriver(q);
   }
   for (; i < TOT_NR_OF_QUEUES; i++) {
      q = printerqs + i;
      q->installed = 1;
      q->qno = i;
      _CpriSetPrinterDriver(q);
   }
   HookCguiDeInit(DeInitPrinter, NULL);
}
