#ifndef SPOOLER_H
#define SPOOLER_H

/*
   The queue concept:

   One queue is fixed to specific output. This output is either of: - A
   physical printer port (index 0-2 for LPT1-LPT3) - Printing to via active
   DOS-driver, actual port determined by current DOS assignement (index 3 or
   QUEUE_DOSPRINT) - A network printer program (index 5 or QUEUE_NETPORT) - Always
   printing to a file (n.u.) - To another queue (to one of queues 0-2).
   (index 6 or QUEUE_OPTIMALPORT) A q contains one file list an one file queue. The
   list contains all currently open files. Normally there is only one in the
   list, but it is allowed with several. The application may have a number of
   files opened for creation. A file that is opened is put at the end of the
   list, but it is not a q - any of the opened files may be closed and then
   immediately moved out from the list. The q contains all files that are
   opened,created and closed by the application, and not yet printed. The
   printing is performed in the background by the spooler, printing ane entry
   at a time from the q. To perform the printing in background the spooler
   used the event-q, printing one character each time (if port is ready). As
   long as there are more characters to print, the spooler will call itself
   via the event-q. If more than one printer are in use the printing will be
   done in parallell. This is conveyd by different calls for differnt qs.

 */
#include <stdio.h>
#ifdef DJGPP
#include <dir.h>
#endif
#ifndef MAXPATH
#define MAXPATH 255
#endif

#include "cpri.h"

#define MAXNETCOMMAND 100
#define PRINTING  1

/* State definitions. */
/* {"Redo", "Avstngd","Ej ansl.","Pap.slut","Offline ","OkntFel"}; */
#define CPR_AVAILABLE   0
#define CPR_PRIOFF      1
#define CPR_NC       2
#define CPR_OOP         3
#define CPR_OFFLINE     4
#define CPR_RANDOM      5
#define CPR_ERROR    6
#define QUEUE_HTML_FILE NR_OF_QUEUES
#define QUEUE_TEXT_FILE (NR_OF_QUEUES+1)
#define TOT_NR_OF_QUEUES (NR_OF_QUEUES+2)

/* The filestruct contains information about one spooler file. A spooler */
/* file contains printer format codes in the ESCAPE-format. A spooler file */
/* is always in a queue. It is however possible to switch q, eiter due to */
/* the "optimal q" concept (i.e. a nonprinting q where the printouts are */
/* waiting for any printing q to be idle) or due to the operator actions in */
/* the monitor window. During printing the ESCAPE-format codes will be */
/* filtered to the codes used by the actual printer which is connected to */
/* the q. */

struct t_qelem;
struct t_printerq;
struct t_driver;

typedef struct t_cpri_vtable {
   struct t_driver *(*Init)(struct t_printerq*);
   void (*Close)(struct t_driver *);
   int (*Driver)(struct t_driver *);
} t_cpri_vtable;

/* A queue data structure. A qstruct will be allocated when the queue is */
/* activated the first time. */
typedef struct t_printerq {
   char net1[MAXNETCOMMAND + 1]; /* Only for the network-printer-port... */
   char net2[MAXNETCOMMAND + 1]; /* Only for the network-printer-port: The
                                    name of the network access program
                                    (name of DOS program file name). The
                                    network-queue do not need to use the
                                    string/control since it sends the entire
                                    file to the network spooler (for this
                                    reason it will be saved to disk). */
   int dostext;                 /* Indicates that the character set (for
                                   national characters) should be converted
                                   from the default "Windows character set"
                                   to the "DOS" (i.e. PC-8 or PC-850) */
   int installed;
   int brand;                   /* Tells which printer is connected to the
                                   port */
   int qno;                     /* The number (index) of the queue. This
                                   number may be used to find the index
                                   number when only a pointer to the q is
                                   available */
   t_cpri_vtable *vtable;       /* Pointer to the functions running the q */
   void (*PrintNextChar)(void*); /* The q processor */
   int hold;                    /* Tells if the queue is in hold-state.
                                   Operator will control that. */
   int busy;                    /* The state of a physical printer: 1=busy
                                   printing a character, 0=not busy */
   int pstate;                  /* Current error state. 0=OK Acronyms above. */
   int prevstate;               /* Old state used to avoid repeating error
                                   messages when the error is persistent */
   int high;                    /* Sequence number used for creating the
                                   spool-file names. */
   FILE *sf;                    /* File currently printing from (NULL if not
                                   printing or if memory buffered
                                   intermediate data). This can be used as
                                   an indication to see if printing is in
                                   progress. (Detection of empty q). */
   char *inbuf;                 /* A pointer to the data if buffered intermediate
                                   data is used, else NULL */
   int files;                   /* Number of entries in the queue */
   int bytes;                   /* Number of bytes left to print in current
                                   file. */
   int c;                       /* Latest read character */
   struct t_qelem *el;          /* A pointer to q elements waiting to be
                                   processed. The first element is the one
                                   currently printing or the one to be
                                   printed next, if not printing. */
   int portaddress;             /* Address to the port that the queue is
                                   assigned to (DOS only) */
   struct t_driver *drv;        /* A pointer to the driver data */
} t_printerq;

struct t_printerq;
extern int _CpriIsPrinting2(void);
extern int _CpriIsPrinting(void);
extern void _CpriSetPrinterDriver(struct t_printerq *q);
extern void _CpriGetAnEntry(struct t_printerq *q);
extern void _CpriFlushPQ(struct t_printerq *q);
extern int _CpriEmptyQ(struct t_printerq *q);
extern char *_CpriProcessedName(struct t_printerq *q);
extern char *_CpriGetDestinationFilename(t_printerq *q);
extern void _CpriCloseQueue(struct t_printerq *q);
extern void _CpriInitBufferedInput(struct t_printerq *q);
extern char *_CpriGetInputPos(struct t_printerq *q);
extern void _CpriSetInputPos(struct t_printerq *q, char *pos);
extern int _CpriGetChar(struct t_printerq *q);
extern void _CpriUnGetChar(struct t_printerq *q);
extern void _CpriGetIntParam(struct t_printerq *q, int *val);
extern void _CpriGetFloatParam(struct t_printerq *q, double *val);
extern void _CpriGetStringParam(struct t_printerq *q,  char *s);
extern int _CpriGetInputSize(struct t_printerq *q);
extern void _CpriPopFront(t_printerq *q);

extern struct t_printerq printerqs[TOT_NR_OF_QUEUES];

#endif
